import unittest

from ether import events

from ether.tests import helpers


class TestABI(unittest.TestCase):

    def test_decode_event(self):
        for i in range(len(helpers.weth_transfers)):
            decoded = events.decode_event(
                helpers.weth_transfers[i],
                helpers.weth_json)
            # print(decoded)
            self.assertEqual(decoded, helpers.weth_parsed[i])

    def test_match_topic0_to_event(self):
        with self.assertRaises(ValueError) as context:
            events.match_topic0_to_event(
                'nope',
                [e for e in helpers.weth_json if e['type'] == 'event'])
        self.assertIn('Topic not found', str(context.exception))

    def test_process_value(self):
        test_vectors = (
            (('address', '0x' + '20' * 32), '0x' + '20' * 20),
            (('bytes', '00' * 32), b'\x00' * 32),
            (('uint', '80' + '00' * 31), 2 ** 255),  # test the sign bit
            (('int', '80' + '00' * 31), -1 * 2 ** 255),  # test the sign bit
            (('bool', '00' * 32), False))
        for v in test_vectors:
            self.assertEqual(
                events.process_value(*v[0]),
                v[1])



import eth_abi

from ether.crypto import keccak256

from typing import Any, cast, Dict, List
from ether.ether_types import EthABI


def make_type_list(function: Dict[str, Any]) -> str:
    '''
    makes a comma-delimted type list
    for use in the signature and encoding
    e.g. '(bytes,bytes,bytes)'
    '''
    return '({})'.format(
        ','.join(t['type'] for t in function['inputs']))


def make_signature(function: Dict[str, Any]) -> str:
    '''
    Makes the function signature
    e.g. transfer(address,value)
    '''
    return '{name}{types}'.format(
        name=function['name'],
        types=make_type_list(function))


def make_selector(function: Dict[str, Any]) -> bytes:
    '''
    makes the 4-byte function selector for encoding data blobs
    '''
    function_signature = make_signature(function)
    return keccak256(function_signature.encode('utf8'))[:4]


def convert_bytes_types(
        function: Dict[str, Any],
        function_args: List[Any]) -> List[Any]:
    '''
    Converts any string-encoded bytes to bytestrings
    Why doesn't eth_abi do this :(
    '''
    converted_args = []
    inputs = function['inputs']
    for i in range(len(function_args)):
        if ('bytes' in inputs[i]['type']
                and type(function_args[i]) is str
                and function_args[i][:2] == '0x'):
            converted_args.append(
                bytes.fromhex(cast(str, function_args[i][2:])))
        else:
            converted_args.append(function_args[i])
    return converted_args


def encode_function_args(
        function: Dict[str, Any],
        function_args: List[Any]) -> bytes:
    '''
    encodes function arguments into a data blob
    This gets prepended with the function selector
    '''
    tmp_args = convert_bytes_types(function, function_args)
    return eth_abi.encode_single(
        make_type_list(function),
        tmp_args)


def matches_args(
        function: Dict[str, Any],
        function_args: List[Any]) -> bool:
    '''
    Checks whether eth_abi will encode each argument with the expected type
    '''
    for i in range(len(function_args)):
        inputs = function['inputs']
        if not eth_abi.is_encodable(inputs[i]['type'], function_args[i]):
            # account for hex-encoded bytes. why doesn't eth_abi do this?
            if ('bytes' not in inputs[i]['type']
                    or type(function_args[i]) is not str
                    or function_args[i][:2] != '0x'):
                return False
    return True


def find_function(
        function_name: str,
        function_args: List[Any],
        abi: EthABI) -> Dict[str, Any]:
    '''
    Find a function within the ABI based on
    a) its name
    b) the number of args
    c) whether the args provided match that function's interface
    '''
    funcs = [e for e in abi if e['type'] == 'function']
    funcs = [f for f in funcs if f['name'] == function_name]
    funcs = [f for f in funcs if len(f['inputs']) == len(function_args)]
    funcs = [f for f in funcs if matches_args(f, function_args)]

    if len(funcs) == 0:
        raise ValueError('no functions with acceptable interface')
    elif len(funcs) != 1:
        raise ValueError('multiple functions with acceptable interface.')
    return funcs[0]


def encode_call(function: Dict[str, Any], function_args: List[Any]) -> bytes:
    '''
    Makes the data blob for a solidity contract function call
    This is a 4 byte selector, and then any number of bytes
    '''
    return (make_selector(function)
            + encode_function_args(function, function_args))


def call(function_name: str, function_args: List[Any], abi: EthABI) -> bytes:
    '''
    Call a function by name
    '''
    function = find_function(function_name, function_args, abi)
    return encode_call(function, function_args)



from typing import Any, Dict, List, Tuple
from mypy_extensions import TypedDict

EthSig = Tuple[int, int, int]

EthABI = List[Dict[str, Any]]


class SignedEthTx(TypedDict):
    to: str
    value: int
    gas: int
    gasPrice: int
    nonce: int
    data: bytes
    v: int
    s: int
    r: int


class UnsignedEthTx(TypedDict):
    to: str
    value: int
    gas: int
    gasPrice: int
    nonce: int
    data: bytes
    chainId: int


class EtherEvent(TypedDict):
    address: str
    blockHash: str
    blockNumber: str
    logIndex: str
    removed: bool
    topics: List[str]
    transactionHash: str
    transactionIndex: str


class UnparsedEtherEvent(EtherEvent):
    data: str


class ParsedEtherEvent(EtherEvent):
    data: Dict[str, Any]



import ssl
import json
import asyncio
import websockets

from ether.ether_types import UnparsedEtherEvent
from typing import cast, Dict, List, Optional, Tuple, Union
from websockets.client import WebSocketClientProtocol


def _encode_int(number: Union[str, int]) -> str:
    try:
        return hex(cast(int, number))
    except TypeError:
        return cast(str, number)


def _id():
    '''Infinite generator for unique request IDs'''
    index = 0
    while 1:
        yield index
        index += 1


URI = 'wss://{network}.infura.io/ws/v3/{project_id}'
_SOCKETS: Dict[str, WebSocketClientProtocol] = {}  # sockets open
_INFLIGHT: Dict[int, asyncio.Future] = {}  # requests awaiting responses
_SUBSCRIPTIONS: Dict[str, asyncio.Queue] = {}  # subscription queues
_IDS = _id()


async def make_socket(
        network: str,
        project_id: str) -> WebSocketClientProtocol:
    '''Sets up a new socket for the network, and registers ping and handler'''
    global _SOCKETS
    if network in _SOCKETS:
        return _SOCKETS[network]
    else:
        uri = URI.format(network=network, project_id=project_id)
        ws = await websockets.connect(uri, ssl=ssl.SSLContext())
        _SOCKETS[network] = ws
        asyncio.ensure_future(_ping(network, project_id, ws))
        asyncio.ensure_future(_handle_incoming(ws))


async def _get_socket(
        network: str) -> WebSocketClientProtocol:
    '''Gets a socket for the network. Blocks if none has been created'''
    while network not in _SOCKETS is None:
        await asyncio.sleep(5)
    return _SOCKETS[network]


async def _ping(network: str, project_id: str, ws) -> None:
    '''
    Periodically pings the websocket to keep it alive
    Attempts to make a new connection if it dies
    '''
    global _SOCKETS
    while True:
        try:
            await ws.ping()
            await asyncio.sleep(15)
        except websockets.exceptions.ConnectionClosed:
            _SOCKETS.pop(network)
            await make_socket(network, project_id)
            break


async def _handle_incoming(ws: WebSocketClientProtocol) -> None:
    '''Router for incoming messages. Distributes them to futures/queues'''
    while True:
        try:
            msg = await ws.recv()
            payload = json.loads(msg)
            # if it has a request ID, it's an RPC response
            if 'error' in payload:
                _INFLIGHT[payload['id']].set_exception(
                    RuntimeError(str(payload['error'])))
                continue
            elif 'id' in payload and payload['id'] in _INFLIGHT:
                _INFLIGHT[payload['id']].set_result(payload['result'])
            # otherwise it's a subscript notification
            # add it to the subscription's queue
            elif 'params' in payload and 'subscription' in payload['params']:
                sub_id = payload['params']['subscription']
                q = _SUBSCRIPTIONS[sub_id]
                await q.put(payload['params']['result'])
        except websockets.exceptions.ConnectionClosed:
            break


async def _RPC(method: str, params: List, network: str):
    '''Internal method for handling RPC calls'''
    # get a new ID and the websket
    req_id = next(_IDS)
    ws = await _get_socket(network)

    # Create a new future store it in our pendings and send it
    fut = asyncio.get_event_loop().create_future()
    _INFLIGHT[req_id] = fut

    # send the request
    payload = {
        'jsonrpc': '2.0',
        'id': req_id,
        'method': method,
        'params': params}
    await ws.send(json.dumps(payload))

    # await the result
    res = await fut
    return res


async def _subscribe(
        params: List,
        network: str) -> Tuple[str, asyncio.Queue]:
    '''Internal method for setting up subscriptions'''
    # set up the queue
    q: asyncio.Queue = asyncio.Queue()

    # get the subscription ID and store the q
    sub_id = await _RPC('eth_subscribe', params, network)
    _SUBSCRIPTIONS[sub_id] = q
    return sub_id, q


async def _fake_subscribe(
        method: str,
        params: List,
        poll_time: int = 10,
        network: str = 'mainnet') -> asyncio.Queue:
    '''Turns an RPC call into a subscription by polling regularly'''
    q: asyncio.Queue = asyncio.Queue()
    task = _fake_sub_poller(q, method, params, poll_time, network)
    asyncio.ensure_future(task)
    return q


async def _fake_sub_poller(
        q: asyncio.Queue,
        method: str,
        params: List,
        poll_time: int = 10,
        network: str = 'mainnet') -> None:
    '''poll for results and add them to the queue'''
    while True:
        try:
            await q.put(await _RPC(method, params, network))
            await asyncio.sleep(poll_time)
        except websockets.exceptions.ConnectionClosed:
            break


async def unsubscribe(sub_ids: List[str], network: str = 'mainnet'):
    '''Cancel a list of subscriptions'''
    return await _RPC('eth_unsubscribe', sub_ids, network)


async def subscribe_to_address_events(
        addresses: List[str],
        topics: Optional[List[str]],
        network: str = 'mainnet') -> Tuple[str, asyncio.Queue]:
    '''Subscribes to event logs at specific addresses'''
    param_dict = {'address': addresses}
    if topics:
        param_dict['topics'] = topics

    return await _subscribe(['logs', param_dict], network)


async def subscribe_to_headers(
        network: str = 'mainnet') -> Tuple[str, asyncio.Queue]:
    '''Subscribes to new headers on the network'''
    return await _subscribe(['newHeads'], network)


async def get_balance(
        address: str,
        block: Union[str, int] = 'latest',
        network: str = 'mainnet') -> int:
    '''Gets the number of wei at an address'''
    res = await _RPC(
        'eth_getBalance',
        [address, _encode_int(block)],
        network)
    return int(res, 16)


async def get_logs(
        address: Optional[str] = None,
        from_block: Union[str, int] = 'earliest',
        to_block: Union[str, int] = 'latest',
        topics: Optional[List[str]] = None,
        blockhash: Optional[str] = None,
        network: str = 'mainnet') -> List[UnparsedEtherEvent]:
    '''Gets logs'''
    params: Dict[str, Union[str, List[str]]] = {}

    if address:
        params['address'] = address
    if topics:
        params['topics'] = topics

    # can't have both a hash and a height range
    if blockhash:
        params['blockhash'] = blockhash
    else:
        params['fromBlock'] = _encode_int(from_block)
        params['toBlock'] = _encode_int(to_block)

    return await _RPC('eth_getLogs', [params], network)


async def get_past_contract_logs(
        address: str,
        topics: Optional[List[str]],
        network: str = 'mainnet') -> List[UnparsedEtherEvent]:
    '''Simpler method to get contract logs'''
    return await get_logs(address=address, topics=topics, network=network)
